#define _GNU_SOURCE
#define _BSD_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#define  E_SCRIPT 127
#define  E_DENIED 126
#define  E_MEMORY 125
#define  E_CONFIG 124
#define  E_USAGE  123

static const char  user[] = USER;
static const char  group[] = GROUP;
static const char  dir[] = PREFIX;

static const char  allow_chars[] = "-.0123456789abcdefghijklmnopqrstuvwxyz";

int main(int argc, char *argv[])
{
    const size_t   dir_len = strlen(dir);
    const uid_t    root_uid = 0;
    const gid_t    root_gid = 0;
    const gid_t    root_gids[1] = { 0 };
    gid_t          allow_gid;
    char          *filename = NULL;

    /* At least one command-line parameter is needed. */
    if (argc < 2 || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")) {
        fprintf(stderr, "\n");
        fprintf(stderr, "Usage: %s [ -h | --help ]\n", argv[0]);
        fprintf(stderr, "       %s SCRIPT [ ARGS... ]\n", argv[0]);
        fprintf(stderr, "\n");
        return E_USAGE;
    }

    /* Obtain the gid of the allowed owner group. */
    {
        struct group *gr = getgrnam(group);
        if (!gr || !gr->gr_name || strcmp(gr->gr_name, group)) {
            fprintf(stderr, "Invalid configuration (no group '%s').\n", group);
            return E_CONFIG;
        }
        allow_gid = gr->gr_gid;
    }

    /* Verify we are run by the proper real user. */
    {
        struct passwd *pw = getpwuid(getuid());
        if (!pw || !pw->pw_name || strcmp(pw->pw_name, user)) {
            fprintf(stderr, "Only user '%s' is allowed to run this.\n", user);
            return E_DENIED;
        }
    }

    /* Verify we are run by the proper real group. */
    if (getgid() != allow_gid) {
        fprintf(stderr, "Only group '%s' is allowed to run this.\n", group);
        return E_DENIED;
    }

    /* Elevate privileges. */
    if (setresgid(root_gid, root_gid, root_gid) ||
        setgroups(1, root_gids) ||
        setresuid(root_uid, root_uid, root_uid)) {
        fprintf(stderr, "No superuser privileges.\n");
        return E_DENIED;
    }

    /* Check the script directory ownership and mode. */
    {
        struct stat  info;

        if (dir_len < 1) {
            fprintf(stderr, "Configuration error (no directory).\n");
            return E_CONFIG;
        }

        if (stat(dir, &info)) {
            fprintf(stderr, "Configuration error (directory not found).\n");
            return E_CONFIG;
        }

        if (info.st_uid != root_uid) {
            fprintf(stderr, "Configuration error (invalid directory owner).\n");
            return E_CONFIG;
        }
        if (info.st_gid != allow_gid) {
            fprintf(stderr, "Configuration error (invalid directory group).\n");
            return E_CONFIG;
        }
        if ((info.st_mode & (S_ISUID | S_ISGID | S_IWGRP | S_IWOTH)) ||
            !(info.st_mode & S_IXUSR)) {
            fprintf(stderr, "Configuration error (invalid directory mode).\n");
            return E_CONFIG;
        }
    }

    /* Construct the path to the file to be executed. */
    {
        const char   *script       = argv[1];
        const size_t  script_len   = (script) ? strlen(script) : 0;
        const size_t  filename_max = dir_len + 1 + script_len + 1;
        char         *p;

        /* Sanity check. */
        if (script_len < 1) {
            fprintf(stderr, "Invalid script name.\n");
            return E_DENIED;
        }

        /* Verify this script name is acceptable. */
        if (strspn(script, allow_chars) != script_len ||
            script[0] == '.' || script[0] == '-' || script[0] == '_') {
            fprintf(stderr, "Invalid script name.\n");
            return E_DENIED;
        }

        filename = malloc(filename_max);
        if (!filename) {
            fprintf(stderr, "Out of memory.\n");
            return E_MEMORY;
        }

        p = filename;

        /* Directory part. */
        memcpy(p, dir, dir_len);
        p += dir_len;

        /* Ensure the directory part ends with a slash. */
        if (p[-1] != '/')
            *(p++) = '/';

        /* File name part. */
        memcpy(p, script, script_len);
        p += script_len;

        /* Terminate the string. */
        *p = '\0';
    }

    /* Verify the target script is owned by root:root, and not writable by others. */
    {
        struct stat  info;

        if (stat(filename, &info)) {
            fprintf(stderr, "%s: %s.\n", filename, strerror(errno));
            return E_SCRIPT;
        }
        if (info.st_uid != root_uid) {
            fprintf(stderr, "Invalid script owner.\n");
            return E_DENIED;
        }
        if (info.st_gid != root_gid) {
            fprintf(stderr, "Invalid script group.\n");
            return E_DENIED;
        }
        if ((info.st_mode & (S_ISUID | S_ISGID | S_IWOTH)) ||
            !(info.st_mode & S_IXUSR)) {
            fprintf(stderr, "Invalid script mode.\n");
            return E_DENIED;
        }
    }

    /* Execute. */
    execv(filename, argv + 1);

    /* Failed. */
    fprintf(stderr, "%s: %s.\n", filename, strerror(errno));
    return E_SCRIPT;
}
